@@ -9,109 +9,8 @@ static const char* FILE_NAME = "Sys/IO/Windows.c";
 
 // These implementations are originally copied form Perl win32/win32.c and win32/win32.h
 
-#include <sys/stat.h>
-#include <unistd.h>
-#include <windows.h>
-#include <errno.h>
-#include <winbase.h>
-#include <fcntl.h>
-#include <direct.h>
-
 #include "Sys-Windows.h"
 
-static OSVERSIONINFO g_osver = {0, 0, 0, 0, 0, ""};
-
-typedef struct {
-    USHORT SubstituteNameOffset;
-    USHORT SubstituteNameLength;
-    USHORT PrintNameOffset;
-    USHORT PrintNameLength;
-    ULONG  Flags;
-    WCHAR  PathBuffer[MAX_PATH*3];
-} MY_SYMLINK_REPARSE_BUFFER;
-
-typedef struct {
-    USHORT SubstituteNameOffset;
-    USHORT SubstituteNameLength;
-    USHORT PrintNameOffset;
-    USHORT PrintNameLength;
-    WCHAR  PathBuffer[MAX_PATH*3];
-} MY_MOUNT_POINT_REPARSE_BUFFER;
-
-typedef struct {
-  ULONG  ReparseTag;
-  USHORT ReparseDataLength;
-  USHORT Reserved;
-  union {
-    MY_SYMLINK_REPARSE_BUFFER SymbolicLinkReparseBuffer;
-    MY_MOUNT_POINT_REPARSE_BUFFER MountPointReparseBuffer;
-    struct {
-      UCHAR DataBuffer[1];
-    } GenericReparseBuffer;
-  } Data;
-} MY_REPARSE_DATA_BUFFER;
-
-#define _S_IFLNK ((unsigned)(_S_IFDIR | _S_IFCHR))
-
-#ifndef EDQUOT			/* Not in errno.h but wanted by POSIX.pm */
-#  define EDQUOT		WSAEDQUOT
-#endif
-
-#define strEQ(string1, string2) (strcmp(string1, string2) == 0)
-#define isSLASH(c) ((c) == '/' || (c) == '\\')
-
-#ifndef SYMBOLIC_LINK_FLAG_DIRECTORY
-#  define SYMBOLIC_LINK_FLAG_DIRECTORY 0x1
-#endif
-
-#ifndef SYMBOLIC_LINK_FLAG_ALLOW_UNPRIVILEGED_CREATE
-#  define SYMBOLIC_LINK_FLAG_ALLOW_UNPRIVILEGED_CREATE 0x2
-#endif
-
-// These are different from Perl's ones, but they must be defined well
-#define PerlDir_mapA(dir) dir
-#define dTHX 
-
-typedef BOOLEAN (__stdcall *pCreateSymbolicLinkA_t)(LPCSTR, LPCSTR, DWORD);
-typedef BOOLEAN (__stdcall *pCreateSymbolicLinkW_t)(LPCWSTR, LPCWSTR, DWORD);
-
-static BOOL
-is_symlink(HANDLE h) {
-    MY_REPARSE_DATA_BUFFER linkdata;
-    const MY_SYMLINK_REPARSE_BUFFER * const sd =
-        &linkdata.Data.SymbolicLinkReparseBuffer;
-    DWORD linkdata_returned;
-
-    if (!DeviceIoControl(h, FSCTL_GET_REPARSE_POINT, NULL, 0, &linkdata, sizeof(linkdata), &linkdata_returned, NULL)) {
-        return FALSE;
-    }
-
-    if (linkdata_returned < offsetof(MY_REPARSE_DATA_BUFFER, Data.SymbolicLinkReparseBuffer.PathBuffer)
-        || (linkdata.ReparseTag != IO_REPARSE_TAG_SYMLINK
-            && linkdata.ReparseTag != IO_REPARSE_TAG_MOUNT_POINT)) {
-        /* some other type of reparse point */
-        return FALSE;
-    }
-
-    return TRUE;
-}
-
-// Same as Perl's is_symlink_name in Win32.c, but call APIs for wide characters(W instead of A).
-static BOOL
-is_symlink_name(const wchar_t *name) {
-    HANDLE f = CreateFileW(name, GENERIC_READ, 0, NULL, OPEN_EXISTING,
-                           FILE_FLAG_OPEN_REPARSE_POINT|FILE_FLAG_BACKUP_SEMANTICS, 0);
-    BOOL result;
-
-    if (f == INVALID_HANDLE_VALUE) {
-        return FALSE;
-    }
-    result = is_symlink(f);
-    CloseHandle(f);
-
-    return result;
-}
-
 // Same as Perl's win32_unlink in Win32.c, but call APIs for wide characters(W instead of A and _w prefixed).
 static int
 win32_unlink(const wchar_t *filename)
@@ -186,161 +85,6 @@ win32_rename(const wchar_t *oname, const wchar_t *newname)
     return 0;
 }
 
-static void
-translate_to_errno(void)
-{
-    /* This isn't perfect, eg. Win32 returns ERROR_ACCESS_DENIED for
-       both permissions errors and if the source is a directory, while
-       POSIX wants EACCES and EPERM respectively.
-    */
-    switch (GetLastError()) {
-    case ERROR_BAD_NET_NAME:
-    case ERROR_BAD_NETPATH:
-    case ERROR_BAD_PATHNAME:
-    case ERROR_FILE_NOT_FOUND:
-    case ERROR_FILENAME_EXCED_RANGE:
-    case ERROR_INVALID_DRIVE:
-    case ERROR_PATH_NOT_FOUND:
-      errno = ENOENT;
-      break;
-    case ERROR_ALREADY_EXISTS:
-      errno = EEXIST;
-      break;
-    case ERROR_ACCESS_DENIED:
-      errno = EACCES;
-      break;
-    case ERROR_PRIVILEGE_NOT_HELD:
-      errno = EPERM;
-      break;
-    case ERROR_NOT_SAME_DEVICE:
-      errno = EXDEV;
-      break;
-    case ERROR_DISK_FULL:
-      errno = ENOSPC;
-      break;
-    case ERROR_NOT_ENOUGH_QUOTA:
-      errno = EDQUOT;
-      break;
-    default:
-      /* ERROR_INVALID_FUNCTION - eg. symlink on a FAT volume */
-      errno = EINVAL;
-      break;
-    }
-}
-
-static int
-do_readlink_handle(HANDLE hlink, char *buf, size_t bufsiz, BOOL *is_symlink) {
-    MY_REPARSE_DATA_BUFFER linkdata;
-    DWORD linkdata_returned;
-
-    if (is_symlink)
-        *is_symlink = FALSE;
-
-    if (!DeviceIoControl(hlink, FSCTL_GET_REPARSE_POINT, NULL, 0, &linkdata, sizeof(linkdata), &linkdata_returned, NULL)) {
-        translate_to_errno();
-        return -1;
-    }
-
-    int bytes_out;
-    BOOL used_default;
-    switch (linkdata.ReparseTag) {
-    case IO_REPARSE_TAG_SYMLINK:
-        {
-            const MY_SYMLINK_REPARSE_BUFFER * const sd =
-                &linkdata.Data.SymbolicLinkReparseBuffer;
-            if (linkdata_returned < offsetof(MY_REPARSE_DATA_BUFFER, Data.SymbolicLinkReparseBuffer.PathBuffer)) {
-                errno = EINVAL;
-                return -1;
-            }
-            bytes_out =
-                WideCharToMultiByte(CP_ACP, WC_NO_BEST_FIT_CHARS,
-                                    sd->PathBuffer + sd->PrintNameOffset/2,
-                                    sd->PrintNameLength/2,
-                                    buf, (int)bufsiz, NULL, &used_default);
-            if (is_symlink)
-                *is_symlink = TRUE;
-        }
-        break;
-    case IO_REPARSE_TAG_MOUNT_POINT:
-        {
-            const MY_MOUNT_POINT_REPARSE_BUFFER * const rd =
-                &linkdata.Data.MountPointReparseBuffer;
-            if (linkdata_returned < offsetof(MY_REPARSE_DATA_BUFFER, Data.MountPointReparseBuffer.PathBuffer)) {
-                errno = EINVAL;
-                return -1;
-            }
-            bytes_out =
-                WideCharToMultiByte(CP_ACP, WC_NO_BEST_FIT_CHARS,
-                                    rd->PathBuffer + rd->PrintNameOffset/2,
-                                    rd->PrintNameLength/2,
-                                    buf, (int)bufsiz, NULL, &used_default);
-            if (is_symlink)
-                *is_symlink = TRUE;
-        }
-        break;
-
-    default:
-        errno = EINVAL;
-        return -1;
-    }
-
-    if (bytes_out == 0 || used_default) {
-        /*() failed.conversion from unicode to ANSI or otherwise() failed.*/
-        errno = EINVAL;
-        return -1;
-    }
-
-    return bytes_out;
-}
-
-static int
-win32_readlink(const char *pathname, char *buf, size_t bufsiz) {
-    if (pathname == NULL || buf == NULL) {
-        errno = EFAULT;
-        return -1;
-    }
-    if (bufsiz <= 0) {
-        errno = EINVAL;
-        return -1;
-    }
-
-    DWORD fileattr = GetFileAttributes(pathname);
-    if (fileattr == INVALID_FILE_ATTRIBUTES) {
-        translate_to_errno();
-        return -1;
-    }
-
-    if (!(fileattr & FILE_ATTRIBUTE_REPARSE_POINT)) {
-        /* not a symbolic link */
-        errno = EINVAL;
-        return -1;
-    }
-
-    HANDLE hlink =
-        CreateFileA(pathname, GENERIC_READ, 0, NULL, OPEN_EXISTING,
-                    FILE_FLAG_OPEN_REPARSE_POINT|FILE_FLAG_BACKUP_SEMANTICS, 0);
-    if (hlink == INVALID_HANDLE_VALUE) {
-        translate_to_errno();
-        return -1;
-    }
-    int bytes_out = do_readlink_handle(hlink, buf, bufsiz, NULL);
-    CloseHandle(hlink);
-    if (bytes_out < 0) {
-        /* errno already set */
-        return -1;
-    }
-
-    if ((size_t)bytes_out > bufsiz) {
-        errno = EINVAL;
-        return -1;
-    }
-
-    return bytes_out;
-}
-
-#define isSLASHW(c) ((c) == L'/' || (c) == L'\\')
-#define strEQW(string1, string2) (wcscmp(string1, string2) == 0)
-
 static int
 win32_symlink(SPVM_ENV* env, SPVM_VALUE* stack, const wchar_t *oldfile, const wchar_t *newfile)
 {
@@ -448,80 +192,6 @@ win32_symlink(SPVM_ENV* env, SPVM_VALUE* stack, const wchar_t *oldfile, const wc
     return 0;
 }
 
-static int win32_lstat(const char* path, struct stat* sbuf)
-{
-  HANDLE f;
-  int result;
-  DWORD attr = GetFileAttributes(path); /* doesn't follow symlinks */
-  
-  if (attr == INVALID_FILE_ATTRIBUTES) {
-    translate_to_errno();
-    return -1;
-  }
-  
-  if (!(attr & FILE_ATTRIBUTE_REPARSE_POINT)) {
-    return stat(path, sbuf);
-  }
-  
-  f = CreateFileA(path, GENERIC_READ, 0, NULL, OPEN_EXISTING,
-                         FILE_FLAG_OPEN_REPARSE_POINT|FILE_FLAG_BACKUP_SEMANTICS, 0);
-  
-  if (f == INVALID_HANDLE_VALUE) {
-    translate_to_errno();
-    return -1;
-  }
-  
-  BOOL is_symlink;
-  
-  int size = do_readlink_handle(f, NULL, 0, &is_symlink);
-  
-  if (!is_symlink) {
-    /* it isn't a symlink, fallback to normal stat */
-    CloseHandle(f);
-    return stat(path, sbuf);
-  }
-  else if (size < 0) {
-    /* some other error, errno already set */
-    CloseHandle(f);
-    return -1;
-  }
-  
-  int32_t fd = _open_osfhandle((intptr_t)f, _O_RDONLY);
-  
-  result = fstat(fd, sbuf);
-  
-  if (result != -1){
-    sbuf->st_mode = (sbuf->st_mode & ~_S_IFMT) | _S_IFLNK;
-    sbuf->st_size = size;
-  }
-  
-  _close(fd);
-  
-  return result;
-}
-
-// Original implementation
-static int win32_realpath(const wchar_t* path, wchar_t* out_path, int32_t out_path_length) {
-  
-  int32_t len = 0; // 0 indicates an error
-  HANDLE hFile = CreateFileW(path, FILE_READ_ATTRIBUTES,
-                    FILE_SHARE_DELETE | FILE_SHARE_READ | FILE_SHARE_WRITE,
-                    NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, NULL);
-  
-  if (hFile == INVALID_HANDLE_VALUE) {
-    len = -1;
-    goto END_OF_FUNC;
-  }
-  
-  len = GetFinalPathNameByHandleW(hFile, out_path, out_path_length, 0);
-  
-  END_OF_FUNC:
-  
-  CloseHandle(hFile);
-  
-  return len;
-}
-
 #endif // _WIN32
 
 int32_t SPVM__Sys__IO__Windows__unlink(SPVM_ENV* env, SPVM_VALUE* stack) {
@@ -599,42 +269,139 @@ int32_t SPVM__Sys__IO__Windows__rename(SPVM_ENV* env, SPVM_VALUE* stack) {
 #endif
 }
 
-int32_t SPVM__Sys__IO__Windows__readlink(SPVM_ENV* env, SPVM_VALUE* stack) {
+int32_t SPVM__Sys__IO__Windows__win_readlink(SPVM_ENV* env, SPVM_VALUE* stack) {
 #if !defined(_WIN32)
-  env->die(env, stack, "Sys::IO::Windows#readlink method is not supported in this system(!defined(_WIN32)).", __func__, FILE_NAME, __LINE__);
+  env->die(env, stack, "Sys::IO::Windows#win_readlink method is not supported in this system(!defined(_WIN32)).", __func__, FILE_NAME, __LINE__);
   return SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_NOT_SUPPORTED_CLASS;
 #else
   int32_t error_id = 0;
-  
   void* obj_path = stack[0].oval;
   if (!obj_path) {
     return env->die(env, stack, "The path $path must be defined.", __func__, FILE_NAME, __LINE__);
   }
   const char* path = env->get_chars(env, stack, obj_path);
   
-  void* obj_buf = stack[1].oval;
-  if (!obj_buf) {
-    return env->die(env, stack, "The buffer $buf must be defined.", __func__, FILE_NAME, __LINE__);
+  wchar_t* path_w = utf8_to_win_wchar(env, stack, path, &error_id, __func__, FILE_NAME, __LINE__);
+  if (error_id) {
+    goto END_OF_FUNC;
   }
-  char* buf = (char*)env->get_chars(env, stack, obj_buf);
-  int32_t buf_length = env->length(env, stack, obj_buf);
-  
-  int32_t bufsiz = stack[2].ival;
-  if (!(bufsiz >= 0)) {
-    return env->die(env, stack, "The buffer size $bufsiz must be greater than or equal to 0.", __func__, FILE_NAME, __LINE__);
-  }
-  if (!(bufsiz <= buf_length)) {
-    return env->die(env, stack, "The buffer size $bufsiz must be less than or equal to the length of the buffer $buf.", __func__, FILE_NAME, __LINE__);
+  DWORD fileattr = GetFileAttributesW(path_w);
+  if (fileattr == INVALID_FILE_ATTRIBUTES) {
+    translate_to_errno();
+    error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
+    goto END_OF_FUNC;
   }
   
-  errno = 0;
-  int32_t placed_length = win32_readlink(path, buf, bufsiz);
-  if (placed_length == -1) {
-    env->die(env, stack, "[System Error]readlink() failed:%s. $path is \"%s\".", env->strerror_nolen(env, stack, errno), path, __func__, FILE_NAME, __LINE__);
-    return SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
+  if (!(fileattr & FILE_ATTRIBUTE_REPARSE_POINT)) {
+    /* not a symbolic link */
+    errno = EINVAL;
+    error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
+    goto END_OF_FUNC;
   }
   
-  stack[0].ival = placed_length;
+  HANDLE hlink =
+    CreateFileW(path_w, GENERIC_READ, 0, NULL, OPEN_EXISTING,
+                FILE_FLAG_OPEN_REPARSE_POINT|FILE_FLAG_BACKUP_SEMANTICS, 0);
+  if (hlink == INVALID_HANDLE_VALUE) {
+    translate_to_errno();
+    error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
+    goto END_OF_FUNC;
+  }
+  
+  
+  MY_REPARSE_DATA_BUFFER linkdata;
+  DWORD linkdata_returned;
+  
+  if (!DeviceIoControl(hlink, FSCTL_GET_REPARSE_POINT, NULL, 0, &linkdata, sizeof(linkdata), &linkdata_returned, NULL)) {
+    translate_to_errno();
+    error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
+    goto END_OF_FUNC;
+  }
+  
+  
+  const wchar_t* PathBuffer = NULL;
+  int32_t PrintNameOffset = -1;
+  int32_t PrintNameLength = -1;
+  switch (linkdata.ReparseTag) {
+    case IO_REPARSE_TAG_SYMLINK: {
+      const MY_SYMLINK_REPARSE_BUFFER * const sd =
+        &linkdata.Data.SymbolicLinkReparseBuffer;
+      if (linkdata_returned < offsetof(MY_REPARSE_DATA_BUFFER, Data.SymbolicLinkReparseBuffer.PathBuffer)) {
+        errno = EINVAL;
+        error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
+        goto END_OF_FUNC;
+      }
+      
+      PathBuffer = sd->PathBuffer;
+      PrintNameOffset = sd->PrintNameOffset;
+      PrintNameLength = sd->PrintNameLength;
+      
+      break;
+    }
+    case IO_REPARSE_TAG_MOUNT_POINT: {
+      const MY_MOUNT_POINT_REPARSE_BUFFER * const rd =
+        &linkdata.Data.MountPointReparseBuffer;
+      if (linkdata_returned < offsetof(MY_REPARSE_DATA_BUFFER, Data.MountPointReparseBuffer.PathBuffer)) {
+        errno = EINVAL;
+        error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
+        goto END_OF_FUNC;
+      }
+      
+      PathBuffer = rd->PathBuffer;
+      PrintNameOffset = rd->PrintNameOffset;
+      PrintNameLength = rd->PrintNameLength;
+      break;
+    }
+    default: {
+      errno = EINVAL;
+      error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
+      goto END_OF_FUNC;
+    }
+  }
+  
+  int32_t bytes_out =
+    WideCharToMultiByte(CP_UTF8, 0,
+                        PathBuffer + PrintNameOffset/2,
+                        PrintNameLength/2,
+                        NULL, 0, NULL, NULL);
+                        
+  if (bytes_out == 0) {
+    /*() failed.conversion from unicode to ANSI or otherwise() failed.*/
+    errno = EINVAL;
+    error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
+    goto END_OF_FUNC;
+  }
+  
+  void* obj_link_text = env->new_string(env, stack, NULL, bytes_out);
+  char* link_text = (char*)env->get_chars(env, stack, obj_link_text);
+  
+  bytes_out =
+    WideCharToMultiByte(CP_UTF8, 0,
+                        PathBuffer + PrintNameOffset/2,
+                        PrintNameLength/2,
+                        link_text, bytes_out, NULL, NULL);
+  if (bytes_out == 0) {
+    /*() failed.conversion from unicode to ANSI or otherwise() failed.*/
+    errno = EINVAL;
+    error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
+    goto END_OF_FUNC;
+  }
+  
+  END_OF_FUNC:
+  
+  if (!(hlink == INVALID_HANDLE_VALUE)) {
+    CloseHandle(hlink);
+  }
+  
+  if (error_id) {
+    if (errno) {
+      env->die(env, stack, "[System Error]win_readlink() failed:%s. $path is \"%s\".", env->strerror_nolen(env, stack, errno), path, __func__, FILE_NAME, __LINE__);
+    }
+    
+    return error_id;
+  }
+  
+  stack[0].oval = obj_link_text;
   
   return 0;
 #endif
@@ -685,44 +452,6 @@ int32_t SPVM__Sys__IO__Windows__symlink(SPVM_ENV* env, SPVM_VALUE* stack) {
 #endif
 }
 
-int32_t SPVM__Sys__IO__Windows__lstat(SPVM_ENV* env, SPVM_VALUE* stack) {
-#if !defined(_WIN32)
-  return env->die(env, stack, "Sys::IO::Windows#lstat method is not supported in this system(!defined(_WIN32)).", __func__, FILE_NAME, __LINE__);
-#else
-
-  int32_t error_id = 0;
-  
-  void* obj_path = stack[0].oval;
-  
-  void* obj_lstat = stack[1].oval;
-  
-  if (!obj_path) {
-    return env->die(env, stack, "The path $path must be defined.", __func__, FILE_NAME, __LINE__);
-  }
-  const char* path = env->get_chars(env, stack, obj_path);
-  
-  if (!obj_lstat) {
-    return env->die(env, stack, "The stat object $lstat must be defined.", __func__, FILE_NAME, __LINE__);
-  }
-  
-  struct stat* stat_buf = env->get_pointer(env, stack, obj_lstat);
-  
-  int32_t status = win32_lstat(path, stat_buf);
-  
-  if (status == -1) {
-    const char* path = env->get_chars(env, stack, obj_path);
-    env->die(env, stack, "[System Error]lstat() failed:%s. $path is \"%s\".", path, env->strerror_nolen(env, stack, errno), path, __func__, FILE_NAME, __LINE__);
-    return SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
-  }
-  
-  stack[0].ival = status;
-  
-  return 0;
-#endif
-
-  return 0;
-}
-
 int32_t SPVM__Sys__IO__Windows__realpath(SPVM_ENV* env, SPVM_VALUE* stack) {
 #if !defined(_WIN32)
   return env->die(env, stack, "Sys::IO::Windows#realpath method is not supported in this system(!defined(_WIN32)).", __func__, FILE_NAME, __LINE__);
@@ -732,51 +461,72 @@ int32_t SPVM__Sys__IO__Windows__realpath(SPVM_ENV* env, SPVM_VALUE* stack) {
   
   void* obj_path = stack[0].oval;
   
-  void* obj_resolved_path = stack[1].oval;
+  void* obj_resolved_path_tmp = stack[1].oval;
   
   if (!obj_path) {
     return env->die(env, stack, "The path $path must be defined.", __func__, FILE_NAME, __LINE__);
   }
   
-  if (obj_resolved_path) {
+  if (obj_resolved_path_tmp) {
     return env->die(env, stack, "The resolved path $resolved_path must not be defined.", __func__, FILE_NAME, __LINE__);
   }
   
   const char* path = env->get_chars(env, stack, obj_path);
   
-  wchar_t* path_w = utf8_to_win_wchar(env, stack, path, &error_id, __func__, FILE_NAME, __LINE__);
+  void* obj_resolved_link_text = NULL;
+  {
+    void* obj_link_text = NULL;
+    stack[0].oval = obj_path;
+    env->call_class_method_by_name(env, stack, "Sys::IO::Windows", "_follow_symlinks_to", 1, &error_id, __func__, FILE_NAME, __LINE__);
+    if (error_id) {
+      goto END_OF_FUNC;
+    }
+    obj_resolved_link_text = stack[0].oval;
+  }
+  const char* resolved_link_text = env->get_chars(env, stack, obj_resolved_link_text);
+  
+  wchar_t* resolved_link_text_w = utf8_to_win_wchar(env, stack, resolved_link_text, &error_id, __func__, FILE_NAME, __LINE__);
   if (error_id) {
     return error_id;
   }
   
-  int32_t len = win32_realpath(path_w, NULL, 0);
+  HANDLE handle = CreateFileW(resolved_link_text_w, FILE_READ_ATTRIBUTES,
+                    FILE_SHARE_DELETE | FILE_SHARE_READ | FILE_SHARE_WRITE,
+                    NULL, OPEN_EXISTING, FILE_FLAG_OPEN_REPARSE_POINT|FILE_FLAG_BACKUP_SEMANTICS, NULL);
   
-  if (len == -1) {
-    env->die(env, stack, "[System Error]CreateFile() failed:the symbolic link is not permitted, broken or not found. $path:\"%s\".", path, __func__, FILE_NAME, __LINE__);
-    return SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
-  }
-  else if (!(len > 0)) {
-    env->die(env, stack, "[System Error]win32_realpath() failed:GetLastError() %d. $path:\"%s\".", GetLastError(), path, __func__, FILE_NAME, __LINE__);
-    return SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
+  if (handle == INVALID_HANDLE_VALUE) {
+    translate_to_errno();
+    error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
+    goto END_OF_FUNC;
   }
   
-  void* obj_resolved_path_w = env->new_short_array(env, stack, len + 1);
+  int32_t needed_len = GetFinalPathNameByHandleW(handle, NULL, 0, 0);
+  
+  if (needed_len == 0) {
+    env->die(env, stack, "[System Error]GetFinalPathNameByHandleW() failed. $path:\"%s\".", path, __func__, FILE_NAME, __LINE__);
+    error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
+    goto END_OF_FUNC;
+  }
+  
+  void* obj_resolved_path_w = env->new_short_array(env, stack, needed_len);
   wchar_t* resolved_path_w = (wchar_t*)env->get_elems_short(env, stack, obj_resolved_path_w);
   
-  len = win32_realpath(path_w, resolved_path_w, len);
+  int32_t len = GetFinalPathNameByHandleW(handle, resolved_path_w, needed_len, 0);
   
-  if (!(len > 0)) {
-    env->die(env, stack, "[System Error]win32_realpath() failed:GetLastError() %d. $path:\"%s\".", GetLastError(), path, __func__, FILE_NAME, __LINE__);
-    return SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
+  if (len == 0) {
+    env->die(env, stack, "[System Error]GetFinalPathNameByHandleW() failed. $path:\"%s\".", path, __func__, FILE_NAME, __LINE__);
+    error_id = SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_SYSTEM_CLASS;
+    goto END_OF_FUNC;
   }
   
-  const char* resolved_path_tmp = win_wchar_to_utf8(env, stack, resolved_path_w, &error_id, __func__, FILE_NAME, __LINE__);
-  
+  char* resolved_path_tmp = (char*)win_wchar_to_utf8(env, stack, resolved_path_w, &error_id, __func__, FILE_NAME, __LINE__);
   if (error_id) {
     return error_id;
   }
-  obj_resolved_path = env->new_string(env, stack, resolved_path_tmp, strlen(resolved_path_tmp));
+  
+  void* obj_resolved_path = env->new_string(env, stack, resolved_path_tmp, strlen(resolved_path_tmp));
   char* resolved_path = (char*)env->get_chars(env, stack, obj_resolved_path);
+  
   int32_t resolved_path_length = env->length(env, stack, obj_resolved_path);
   
   if (strncmp(resolved_path, "\\\\?\\", 4) == 0) {
@@ -798,8 +548,50 @@ int32_t SPVM__Sys__IO__Windows__realpath(SPVM_ENV* env, SPVM_VALUE* stack) {
     }
   }
   
+  END_OF_FUNC:
+  
+  if (!(handle == INVALID_HANDLE_VALUE)) {
+    CloseHandle(handle);
+  }
+  
+  if (error_id) {
+    if (errno) {
+      env->die(env, stack, "[System Error]win_realpath() failed:%s. $path:\"%s\".", env->strerror_nolen(env, stack, errno), path, __func__, FILE_NAME, __LINE__);
+    }
+    
+    return error_id;
+  }
+  
   stack[0].oval = obj_resolved_path;
   
   return 0;
 #endif
 }
+
+int32_t SPVM__Sys__IO__Windows__is_symlink(SPVM_ENV* env, SPVM_VALUE* stack) {
+#if !defined(_WIN32)
+  env->die(env, stack, "Sys::IO::Windows#is_symlink method is not supported in this system(!defined(_WIN32)).", __func__, FILE_NAME, __LINE__);
+  return SPVM_NATIVE_C_BASIC_TYPE_ID_ERROR_NOT_SUPPORTED_CLASS;
+#else
+  
+  int32_t error_id = 0;
+  
+  void* obj_path = stack[0].oval;
+  if (!obj_path) {
+    return env->die(env, stack, "The path $path must be defined.", __func__, FILE_NAME, __LINE__);
+  }
+  
+  const char* path = env->get_chars(env, stack, obj_path);
+  
+  wchar_t* path_w = utf8_to_win_wchar(env, stack, path, &error_id, __func__, FILE_NAME, __LINE__);
+  if (error_id) {
+    return error_id;
+  }
+  
+  int32_t ret = is_symlink_name(path_w);
+  
+  stack[0].ival = ret;
+  
+  return 0;
+#endif
+}
